/*
 *  channel.c: channel management
 *
 *  Copyright (C) 2003 Stuart Walsh
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: channels.c,v 1.8 2003/11/05 00:33:42 stu Exp $
 *
 * Original copyright follows:
 *
 * Services is copyright (c) 1996-1999 Andrew Church.
 *     E-mail: <achurch@dragonfire.net>
 * Services is copyright (c) 1999-2000 Andrew Kempe.
 *     E-mail: <theshadow@shadowfire.org>
 */

#include "stdinc.h"
#include "tools.h"
#include "channel.h"
#include "chanserv.h"
#include "memory.h"
#include "user.h"
#include "send.h"
#include "log.h"
#include "mystring.h"
#include "hash.h"
#include "misc.h"
#include "smysql.h"

static void cmode_simple(struct Channel*, unsigned int, short int);
static void cmode_nosuch(struct Channel*, unsigned int, short int);

static void
cmode_nosuch(struct Channel *c, unsigned int mode, short int dir)
{
}

static void
cmode_simple(struct Channel *c, unsigned int mode, short int dir)
{
}

struct chanmodes
{
    void *param;
    void (*func) (struct Channel *chan, unsigned int mode, short int dir);
    
} mode_table[255] = 
{
    { NULL, cmode_nosuch }, 
    { NULL, cmode_nosuch }, /* A */
    { NULL, cmode_nosuch }, /* B */
    { NULL, cmode_nosuch }, /* C */
    { NULL, cmode_nosuch }, /* D */
    { NULL, cmode_nosuch }, /* E */
    { NULL, cmode_nosuch }, /* F */
    { NULL, cmode_nosuch }, /* E */
    { NULL, cmode_nosuch }, /* G */
    { NULL, cmode_nosuch }, /* H */
    { NULL, cmode_nosuch }, /* I */
    { NULL, cmode_nosuch }, /* J */
    { NULL, cmode_nosuch }, /* K */
    { NULL, cmode_nosuch }, /* L */
    { NULL, cmode_nosuch }, /* M */
    { NULL, cmode_nosuch }, /* N */
    { NULL, cmode_nosuch }, /* O */
    { NULL, cmode_nosuch }, /* P */
    { NULL, cmode_nosuch }, /* Q */
    { NULL, cmode_nosuch }, /* R */
    { NULL, cmode_nosuch }, /* S */
    { NULL, cmode_nosuch }, /* T */
    { NULL, cmode_nosuch }, /* U */
    { NULL, cmode_nosuch }, /* V */
    { NULL, cmode_nosuch }, /* W */
    { NULL, cmode_nosuch }, /* X */
    { NULL, cmode_nosuch }, /* Y */
    { NULL, cmode_nosuch }, /* Z */
    { NULL, cmode_nosuch },
    { NULL, cmode_nosuch },
    { NULL, cmode_nosuch },
    { NULL, cmode_nosuch },
    { NULL, cmode_nosuch },
    { NULL, cmode_nosuch },
    { NULL, cmode_nosuch }, /* a */
    { NULL, cmode_nosuch }, /* b */
    { (void*)CMODE_NOCOLOURS, cmode_simple }, /* c */
    { NULL, cmode_nosuch }, /* d */
    { NULL, cmode_nosuch }, /* e */
    { NULL, cmode_nosuch }, /* f */
    { NULL, cmode_nosuch }, /* g */
    { NULL, cmode_nosuch }, /* h */
    { (void*)CMODE_INVITE, cmode_simple }, /* i */
    { NULL, cmode_nosuch }, /* j */
    { NULL, cmode_nosuch }, /* k */
    { NULL, cmode_nosuch }, /* l */
    { (void*)CMODE_MODERATED, cmode_simple }, /* m */
    { (void*)CMODE_NOEXTERNALMSG, cmode_simple }, /* n */
    { NULL, cmode_nosuch }, /* o */
    { (void*)CMODE_PRIVATE, cmode_simple }, /* p */
    { NULL, cmode_nosuch }, /* q */
    { NULL, cmode_nosuch }, /* r */
    { (void*)CMODE_SECRET , cmode_simple }, /* s */
    { (void*)CMODE_TOPICOPS, cmode_simple }, /* t */
    { NULL, cmode_nosuch }, /* u */
    { NULL, cmode_nosuch }, /* v */
    { NULL, cmode_nosuch }, /* w */
    { NULL, cmode_nosuch }, /* x */
    { NULL, cmode_nosuch }, /* y */
    { NULL, cmode_nosuch }, /* z */
    { 0, '\0' },
};

struct mode_letter flags[] = 
{
    { CMODE_NOCOLOURS,      'c' },
    { CMODE_INVITE,         'i' },
    { CMODE_MODERATED,      'm' },
    { CMODE_NOEXTERNALMSG,  'n' },
    { CMODE_PRIVATE,        'p' },
    { CMODE_SECRET,         's' },
    { CMODE_TOPICOPS,       't' },
    { 0, '\0' }
};

dlink_list channel_list;

static void mysql_init_channel(struct Channel *c);
static void mysql_destroy_channel(struct Channel *c);

/* Return statistics.  Pointers are assumed to be valid. */

void 
get_channel_stats(long *nrec, long *memuse)
{
	long count, mem;
	size_t j;
	struct Channel *chan;
    dlink_node *ptr;

	count = 0;
	mem = 0;

    DLINK_FOREACH(ptr, channel_list.head)
    {
        chan = ptr->data;
        count++;
        mem += sizeof(*chan);
        if (chan->key)
            mem += strlen(chan->key) + 1;
        mem += sizeof(char *) * chan->bansize;

        for (j = 0; j < chan->bancount; j++) 
        {
            if (chan->newbans[j]) 
            {
                if (chan->newbans[j]->mask)
                    mem += strlen(chan->newbans[j]->mask) + 1;
                mem += sizeof(chan->newbans[j]);
            }
        }
       
        mem += sizeof(struct User) * dlink_list_length(&chan->users);
        mem += sizeof(struct User) * dlink_list_length(&chan->chanops);
        mem += sizeof(struct User) * dlink_list_length(&chan->voices);
    }
    *nrec = count;
    *memuse = mem;
}

#ifdef DEBUG_COMMANDS

/* Send the current list of channels to the named user. */

void 
send_channel_list(struct User *user)
{
	struct Channel *c;
    struct User *u, *u2;
    register const struct mode_letter *cmode;
	char s[16];
    char buf[BUFSIZE], modebuf[BUFSIZE];
    char *end, *tmp;
	int isop, isvoice;
	const char *source = user->nick;
    dlink_node *ptr, *ptr2;

    DLINK_FOREACH(ptr, channel_list.head)
    {
        c = ptr->data;
		snprintf(s, sizeof(s), " %d", c->limit);
        tmp = modebuf;

        for(cmode = flags; cmode->letter != '\0'; cmode++)
        {
            if(c->mode & cmode->mode)
            {
                *tmp++ = cmode->letter;
                if(cmode->mode == CMODE_KEY)
                {
                    char *p;

                    *tmp++ = ' ';
                    for(p = c->key; *p; p++)
                        *tmp++ = *p;
                }
                else if(cmode->mode == CMODE_LIMIT)
                {
                    char tmp2[BUFSIZE];
                    char *p;

                    *tmp++ = ' ';
                    snprintf(tmp2, BUFSIZE, "%d", c->limit);
                    for(p = tmp2; *p; p++)
                        *tmp++ = *p;
                }
            }
        }
		notice(s_OperServ, source, "%s %lu +%s %s", c->name, c->creation_time, 
                tmp, c->topic ? c->topic : "");
		end = buf;
		end += snprintf(end, sizeof(buf) - (end - buf), "%s", c->name);
		DLINK_FOREACH(ptr, c->users.head)
        {
            u = ptr->data;
			isop = isvoice = 0;
            DLINK_FOREACH(ptr2, c->chanops.head)
            {
                u2 = ptr2->data;
				if (u == u2) 
                {
					isop = 1;
					break;
				}
			}
            DLINK_FOREACH(ptr2, c->voices.head)
            {
                u2 = ptr->data;
				if (u == u2) 
                {
					isvoice = 1;
					break;
				}
			}
            /* XXX HAlfops */
			end += snprintf(end, sizeof(buf) - (end - buf), " %s%s%s", 
                    isvoice ? "+" : "", isop ? "@" : "", u->nick);
		}
        notice(s_OperServ, source, buf);
    }
}


/* Send list of users on a single channel, taken from strtok(). */

void 
send_channel_users(struct User *user)
{
    char *chan = strtok(NULL, " ");
    struct Channel *c;
    struct User *u;
    const char *source = user->nick;
    dlink_node *ptr;

    if(chan == NULL)
        return;

    c = findchan(chan);
    
    if (c == NULL) 
    {
        notice(s_OperServ, source, "Channel %s not found!", chan);
        return;
    }
    notice(s_OperServ, source, "Channel %s users:", chan);

    DLINK_FOREACH(ptr, c->users.head)
    {
        u = ptr->data;
        notice(s_OperServ, source, "%s", u->nick);
    }
    notice(s_OperServ, source, "Channel %s chanops:", chan);
    
    DLINK_FOREACH(ptr, c->chanops.head)
    {
        u = ptr->data;
        notice(s_OperServ, source, "%s", u->nick);
    }
    notice(s_OperServ, source, "Channel %s voices:", chan);
    
    DLINK_FOREACH(ptr, c->voices.head)
    {
        u = ptr->data;
        notice(s_OperServ, source, "%s", u->nick);
    }
}

#endif				/* DEBUG_COMMANDS */

/* Return the struct Channel structure corresponding to the named channel, or NULL
 * if the channel was not found.  chan is assumed to be non-NULL and valid
 * (i.e. pointing to a channel name of 2 or more characters). */

struct Channel *
findchan(const char *chan)
{
    if(chan == NULL)
        return NULL;
    return(hash_find_channel(chan));
}

/*
 * Return the struct Channel structure corresponding to the given channel_id, or NULL
 * if the channel was not found.  channel_id is assumed to be non-zero.
 */
struct Channel *
findchan_by_id(unsigned int channel_id)
{
    struct Channel *c;
    dlink_node *ptr;

    if (debug >= 3)
        slog("debug: findchan_by_id(%u)", channel_id);

    DLINK_FOREACH(ptr, channel_list.head)
    {
        c = ptr->data;
        if (c->channel_id == channel_id)
            return(c);
    }

    return(NULL);
}


/* Add/remove a user to/from a channel, creating or deleting the channel as
 * necessary.  If creating the channel, restore mode lock and topic as
 * necessary.  Also check for auto-opping and auto-voicing. */

void 
chan_adduser(struct User *user, const char *chan)
{
    struct Channel *c = findchan(chan);

    if (c == NULL) 
    {
        if (debug)
            slog("debug: Creating channel %s", chan);
        /* Allocate pre-cleared memory */
        c = smalloc(sizeof(struct Channel));
        strscpy(c->name, chan, sizeof(c->name));
        hash_add_channel(c);
        c->creation_time = time(NULL);
        /* Store channel ID in channel record */
        c->channel_id = mysql_findchan(chan);

        mysql_init_channel(c);
        /* Restore locked modes and saved topic */
        check_modes(chan);
        restore_topic(chan);

        if(c->floodserv)
            send_cmd(ServerName, "SJOIN %lu %s + :%s", time(NULL), c->name, 
                    s_FloodServ);
    }

    if (check_should_op(user, chan)) 
    {
        dlinkAdd(user, &user->cnode, &c->chanops);
    } 
    else if (check_should_voice(user, chan)) 
    {
        dlinkAdd(user, &user->cnode, &c->voices);
    }
    /* XXX halfops */

    dlinkAdd(user, &user->cnode, &c->users);
}


/* FIXME: I think this is where all the problems relating to the AKICK ENFORCE
 * coredumps are occuring. Remove debug code when things are back to normal.
 *
 * Previously, we bailed out of this function if the user was not in the user
 * list. Now we go through the entire thing to make sure we remove them from
 * the op and voice lists. This raises issues about the code that handles
 * these lists in the first place - why is it possible for users to be in op
 * and voice lists when they're not in the user list?
 * -TheShadow
 */

void 
chan_deluser(struct User * user, struct Channel * c)
{
    struct User *u;
    size_t i;
    dlink_node *ptr;

    if (debug)
        slog("channel: chan_deluser() called...");

    DLINK_FOREACH(ptr, c->users.head)
    {
        u = ptr->data;
        if(u == user)
        {
            dlinkDelete(&u->cnode, &c->users);
            break;
        } 

    }
    

    DLINK_FOREACH(ptr, c->chanops.head)
    {
        u = ptr->data;
        if(u == user)
        {
            dlinkDelete(&u->cnode, &c->chanops);
            break;
        }
    }

    DLINK_FOREACH(ptr, c->voices.head)
    {
        u = ptr->data;
        if(u == user)
        {
            dlinkDelete(&u->cnode, &c->voices);
            break;
        }
    }


    if (dlink_list_length(&c->users) == 0) 
    {
        if (debug)
            slog("debug: Deleting channel %s", c->name);

        mysql_destroy_channel(c);

        if (c->key)
            sfree(c->key);
        for (i = 0; i < c->bancount; ++i) 
        {
            if (c->newbans[i]) 
            {
                if (c->newbans[i]->mask)
                    sfree(c->newbans[i]->mask);
                sfree(c->newbans[i]);
            } 
            else
                slog("channel: BUG freeing %s: bans[%d] is NULL!", c->name, i);
        }
        if (c->bansize)
            sfree(c->newbans);

        hash_del_channel(c);
        sfree(c);
    } 
    else 
        check_autolimit(c, time(NULL), 0);

    if (debug)
        slog("channel: chan_deluser() complete.");
}

/* Handle a channel MODE command. */

void 
do_cmode(const char *source, int ac, char **av)
{
    /* 31 modes + leeway. */
    static char lastmodes[40];
    int dir = CMODE_ADD;
    struct Channel *chan;
    //struct Ban **b;
    char c;
    char *modestr;

    modestr = av[1];
    chan = findchan(av[0]);

    if (chan == NULL) 
    {
        slog("channel: MODE %s for nonexistent channel %s",
                merge_args(ac - 1, av + 1), av[0]);
        return;
    }

    /* This shouldn't trigger on +o, etc. */
    if (strchr(source, '.') && strcmp(source, ServerName) != 0 &&
            !modestr[strcspn(modestr, "bov")]) 
    {

        if (time(NULL) != chan->server_modetime || 
                strcmp(modestr, lastmodes) != 0) 
        {
            chan->server_modecount = 0;
            chan->server_modetime = time(NULL);
            strscpy(lastmodes, modestr, sizeof(lastmodes));
        }
        chan->server_modecount++;
    }

    ac -= 2;
    av += 2;

    for(; (c = *modestr) != '\0'; modestr++) 
    {
        int tableposition;
        
        switch (c) 
        {
            case '+':
                dir = CMODE_ADD;
                break;
            case '-':
                dir = CMODE_DEL;
                break;
            default:
                if(c < 'A' || c > 'z')
                    tableposition = 0;
                else
                    tableposition = c - 'A' + 1;
                mode_table[tableposition].func(chan, c, dir);
                break;
        }
    }
#if 0
            case 'i':
                if (add)
                    chan->mode |= CMODE_i;
                else
                    chan->mode &= ~CMODE_i;
                break;
            case 'm':
                if (add)
                    chan->mode |= CMODE_m;
                else
                    chan->mode &= ~CMODE_m;
                break;
            case 'M':
                if (add)
                    chan->mode |= CMODE_M;
                else
                    chan->mode &= ~CMODE_M;
                break;
            case 'n':
                if (add)
                    chan->mode |= CMODE_n;
                else
                    chan->mode &= ~CMODE_n;
                break;
            case 'p':
                if (add)
                    chan->mode |= CMODE_p;
                else
                    chan->mode &= ~CMODE_p;
                break;
            case 's':
                if (add)
                    chan->mode |= CMODE_s;
                else
                    chan->mode &= ~CMODE_s;
                break;
            case 't':
                if (add)
                    chan->mode |= CMODE_t;
                else
                    chan->mode &= ~CMODE_t;
                break;
            case 'c':
                if (add)
                    chan->mode |= CMODE_c;
                else
                    chan->mode &= ~CMODE_c;
                break;
            case 'O':
                if (add)
                    chan->mode |= CMODE_O;
                else
                    chan->mode &= ~CMODE_O;
                break;
            case 'R':
                if (add)
                    chan->mode |= CMODE_R;
                else
                    chan->mode &= ~CMODE_R;
                break;
#if 0
            case 'r':
                if (add)
                    chan->mode |= CMODE_r;
                else
                    chan->mode &= ~CMODE_r;
                break;
#endif
            case 'k':
                if (--ac < 0) 
                {
                    slog("channel: MODE %s %s: missing parameter for %ck", 
                            chan->name, modestr, add ? '+' : '-');
                    break;
                }
                if (chan->key) 
                {
                    free(chan->key);
                    chan->key = NULL;
                }
                if (add) 
                {
                    chan->key = sstrdup(*av);
                    av++;
                }
                break;
            case 'l':
                if (add) 
                {
                    if (--ac < 0) 
                    {
                        slog("channel: MODE %s %s: missing parameter for +l", 
                                chan->name, modestr);
                        break;
                    }
                    chan->limit = strtoul(*av, NULL, 10);
                    av++;
                } 
                else 
                    chan->limit = 0;
                break;
            case 'b':
                if (--ac < 0) 
                {
                    slog("channel: MODE %s %s: missing parameter for %cb", 
                            chan->name, modestr, add ? '+' : '-');
                    break;
                }
                if (add) 
                {
                    if (chan->bancount >= chan->bansize) 
                    {
                        chan->bansize += 8;
                        chan->newbans = srealloc(chan->newbans, 
                                sizeof(Ban *) * chan->bansize);
                    }
                    chan->newbans[chan->bancount] = smalloc(sizeof(Ban));
                    chan->newbans[chan->bancount]->mask = sstrdup(*av++);
                    chan->newbans[chan->bancount]->time = time(NULL);
                    chan->bancount++;
                } 
                else 
                {
                    b = chan->newbans;
                    i = 0;

                    while (i < chan->bancount && strcmp((*b)->mask, *av) != 0) 
                    {
                        i++;
                        b++;
                    }
                    if (i < chan->bancount) 
                    {
                        chan->bancount--;
                        if (i < chan->bancount) 
                        {
                            free((*b)->mask);
                            memmove(b, b + 1, 
                                    sizeof(Ban *) * (chan->bancount - i));
                        }
                    }
                    av++;
                }
                break;
            case 'o':
                if (--ac < 0) 
                {
                    slog("channel: MODE %s %s: missing parameter for %co", 
                            chan->name, modestr, add ? '+' : '-');
                    break;
                }
                nick = *av++;
                if (add) 
                {
                    for (u = chan->chanops; 
                            u && stricmp(u->user->nick, nick) != 0; u = u->next);
                    if (u)
                        break;
                    user = finduser(nick);
                    if (user == NULL) 
                    {
                        slog("channel: MODE %s +o for nonexistent user %s",
                                chan->name, nick);
                        break;
                    }
                    if (debug)
                        slog("debug: Setting +o on %s for %s", chan->name, nick);
                    if (!check_valid_op(user, chan->name, !!strchr(source, '.')))
                        break;
                    u = smalloc(sizeof(*u));
                    u->next = chan->chanops;
                    u->prev = NULL;
                    if (chan->chanops)
                        chan->chanops->prev = u;
                    chan->chanops = u;
                    u->user = user;
                } 
                else 
                {
                    for (u = chan->chanops; u && stricmp(u->user->nick, nick); 
                            u = u->next);
                    if (u == NULL)
                        break;
                    if (u->next)
                        u->next->prev = u->prev;
                    if (u->prev)
                        u->prev->next = u->next;
                    else
                        chan->chanops = u->next;
                    free(u);
                }
                break;
            case 'v':
                if (--ac < 0) 
                {
                    slog("channel: MODE %s %s: missing parameter for %cv", 
                            chan->name, modestr, add ? '+' : '-');
                    break;
                }
                nick = *av++;
                if (add) 
                {
                    for (u = chan->voices; u && stricmp(u->user->nick, nick); 
                            u = u->next);
                    if (u != NULL)
                        break;
                    user = finduser(nick);
                    if (user == NULL) 
                    {
                        slog("channe: MODE %s +v for nonexistent user %s",
                                chan->name, nick);
                        break;
                    }
                    if (debug)
                        slog("debug: Setting +v on %s for %s", chan->name, nick);
                    u = smalloc(sizeof(*u));
                    u->next = chan->voices;
                    u->prev = NULL;
                    if (chan->voices)
                        chan->voices->prev = u;
                    chan->voices = u;
                    u->user = user;
                } 
                else 
                {
                    for (u = chan->voices; u && stricmp(u->user->nick, nick);
                            u = u->next) ;
                    if (u == NULL)
                        break;
                    if (u->next)
                        u->next->prev = u->prev;
                    if (u->prev)
                        u->prev->next = u->next;
                    else
                        chan->voices = u->next;
                    free(u);
                }
                break;
        }		/* switch */
    } /* while (*s) */
#endif

    /* Check modes against ChanServ mode lock */
    check_modes(chan->name);
}

/* Handle a TOPIC command. */

void 
do_topic(const char *nick, int ac, char **av)
{
    struct Channel *c = findchan(av[0]);

    if (c == NULL) 
    {
        slog("channel: TOPIC %s for nonexistent channel %s", 
                merge_args(ac - 1, av + 1), av[0]);
        return;
    }
    if (check_topiclock(av[0], nick, c))
        return;
    strscpy(c->topic_setter, nick, sizeof(c->topic_setter));
    c->topic_time = time(NULL);
    if (ac > 1 && *av[1])
    {
        strncpy(c->topic, av[1], TOPICLEN);
        c->topic[TOPICLEN] = '\0';
    }
    record_topic(av[0]);
}

/*
 * Go through every channel that is currently active and has a limit
 * offset or ban clearance time.  If it's more than ci->limit_tolerance
 * off from the current user count and it's beyond the time period for
 * updates, then set the limit.  Then check the ban list for any that need
 * expiring.
 *  -grifferz
 */

void 
chan_update_events(time_t now)
{
    time_t threshold;
    size_t i;
    unsigned int count;
    char *av[3];
    struct Channel *chan;
    struct Ban **to_remove;
    dlink_node *ptr;

    DLINK_FOREACH(ptr, channel_list.head)
    {
        chan = ptr->data;
        if (!chan->channel_id) 
        {
            /* Channel isn't registered. */
            continue;
        }

        /* Check for channels that need their limit changed. */
        if (chan->limit_offset > 0 &&
                (unsigned)(now - chan->last_limit_time) >= 
                (chan->limit_period * 60)) 
        {
            check_autolimit(chan, now, 1);
        }

        /* Check for bans that need to be removed. */
        if (chan->bantime) 
        {
            to_remove = NULL;

            /*
             * Any ban set before this time needs to be
             * removed.
             */
            threshold = now - (chan->bantime * 60);

            /*
             * First find the ones to remove - we won't remove
             * them till later because doing so would require a
             * separate copy of the banlist and it's wasteful
             * to take a copy every time unless we know we're
             * going to remove something.
             */
            for (i = 0, count = 0; i < chan->bancount; i++) 
            {
                if (chan->newbans[i]->time < threshold) 
                {
                    /* Time to remove this one. */
                    to_remove = srealloc(to_remove,
                            sizeof(*to_remove) *
                            (count + 1));
                    to_remove[count] =
                        smalloc(sizeof(struct Ban));
                    to_remove[count]->mask =
                        sstrdup(chan->newbans[i]->mask);
                    count++;
                }
            }

            if (count) 
            {
                av[0] = chan->name;
                av[1] = sstrdup("-b");

                for (i = 0; i < count; i++) 
                {
                    av[2] = sstrdup(to_remove[i]->mask);
                    send_cmd(s_ChanServ,  "MODE %s -b %s", 
                            av[0], av[2]);
                    do_cmode(s_ChanServ, 3, av);
                    sfree(av[2]);
                    sfree(to_remove[i]->mask);
                    sfree(to_remove[i]);
                } /* for count */
                sfree(av[1]);
                sfree(to_remove);
            } /* if count */
        } /* if bantime */
    } /* DLINK_FOREACH */
}

/*
 * Count the number of channels whose name matches a POSIX regular
 * expression.
 */
unsigned int 
chan_count_regex_match(regex_t *r)
{
    unsigned int cnt;
    struct Channel *c;
    dlink_node *ptr;

    cnt = 0;

    DLINK_FOREACH(ptr, channel_list.head)
    {
        c = ptr->data;
        if (regexec(r, c->name, 0, NULL, 0) == 0)
            cnt++;
    }
    return(cnt);
}

/*
 * Initialise various parts of struct Channel structure from SQL data.
 */
static void 
mysql_init_channel(struct Channel *c)
{
    MYSQL_ROW row;
    unsigned int fields, rows;
    MYSQL_RES *result;

    if (c == NULL)
        return;

    c->last_limit_time = 0;
    c->limit_offset = 0;
    c->limit_tolerance = 0;
    c->limit_period = 0;
    c->bantime = 0;

    if (!c->channel_id)
        return;

    result = smysql_bulk_query(mysqlconn, &fields, &rows,
            "SELECT last_limit_time, limit_offset, limit_tolerance, "
            "limit_period, bantime, floodserv_protected FROM channel "
            " WHERE channel_id=%u", c->channel_id);

    if ((row = smysql_fetch_row(mysqlconn, result))) 
    {
        c->last_limit_time = atoi(row[0]);
        c->limit_offset = atoi(row[1]);
        c->limit_tolerance = atoi(row[2]);
        c->limit_period = atoi(row[3]);
        c->bantime = atoi(row[4]);
        c->floodserv = atoi(row[5]);
    }
    mysql_free_result(result);
}

/*
 * Store channel data as last user leaves the channel.
 */
static void 
mysql_destroy_channel(struct Channel *c)
{
    unsigned int fields, rows;
    MYSQL_RES *result;

    if (c == NULL || !c->channel_id)
        return;

    result = smysql_bulk_query(mysqlconn, &fields, &rows,
            "UPDATE channel SET last_limit_time=%lu, limit_offset=%u, "
            "limit_tolerance=%u, limit_period=%u, bantime=%u "
            "WHERE channel_id=%u", c->last_limit_time, c->limit_offset,
            c->limit_tolerance, c->limit_period, c->bantime, c->channel_id);
    mysql_free_result(result);
}

/*
 * Check if a specific channel's autolimit must be updated.  If allow_raise is
 * 0 then the limit can never go up.  This is used when calling this function
 * after a user leaves a channel (PART, QUIT, KICK etc).
 */
void 
check_autolimit(struct Channel *chan, time_t now, int allow_raise)
{
    unsigned int fields, rows, usercnt;
    int delta;
    MYSQL_RES *result;

    if (!(chan && chan->channel_id && chan->limit_offset))
        return;

    /*
     * Never allow channel limit changes more often than every 10 secs,
     * otherwise you can end up with a flood of them in situations like
     * netsplit.
     */
    if (now - chan->last_limit_time <= 10)
        return;

    usercnt = dlink_list_length(&chan->users);

    delta = usercnt + chan->limit_offset - chan->limit;

    /*
     * We want to change the limit if it needs to go lower, or
     * if it is high enough to go past the tolerance.
     */
    if (delta < 0 || (((unsigned)delta > chan->limit_tolerance) && allow_raise)) 
    {
        chan->limit = usercnt + chan->limit_offset;
        send_cmd(s_ChanServ, "MODE %s +l %d", chan->name, chan->limit);
    }

    chan->last_limit_time = now;

    result = smysql_bulk_query(mysqlconn, &fields, &rows,
            "UPDATE channel SET last_limit_time=%lu WHERE channel_id=%u",
            chan->last_limit_time, chan->channel_id);
    mysql_free_result(result);
}

