/* Routines for time-delayed actions.
 *
 * Blitzed Services copyright (c) 2000-2002 Blitzed Services team
 *     E-mail: services@lists.blitzed.org
 * Based on ircservices-4.4.8:
 * copyright (c) 1996-1999 Andrew Church.
 *     E-mail: <achurch@dragonfire.net>
 * copyright (c) 1999-2000 Andrew Kempe.
 *     E-mail: <theshadow@shadowfire.org>
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "stdinc.h"
#include "tools.h"
#include "user.h"
#include "services.h"
#include "timeout.h"
#include "send.h"
#include "log.h"
#include "memory.h"

/* $Id: timeout.c,v 1.4 2003/11/05 00:33:42 stu Exp $ */

dlink_list timeout_list;

#ifdef DEBUG_COMMANDS

/* Send the timeout list to the given user. */

void 
send_timeout_list(struct User *u)
{
    struct Timeout *to;
    dlink_node *ptr;

    notice(s_OperServ, u->nick, "Now: %ld", time(NULL));
    DLINK_FOREACH(ptr, timeout_list.head)
    {
        to = ptr->data;
        notice(s_OperServ, u->nick, "%p: %ld: %p (%p)",
                to, to->timeout, to->code, to->data);
    }
}

#endif	/* DEBUG_COMMANDS */

/* Check the timeout list for any pending actions. */

void
check_timeouts(void)
{
    struct Timeout *to;
    time_t t = time(NULL);
    dlink_node *ptr, *ptr2;

    if (debug >= 2)
	slog("debug: Checking timeouts at %ld", t);

    DLINK_FOREACH_SAFE(ptr, ptr2, timeout_list.head)
    {
        to = ptr->data;
        if (t < to->timeout) 
            continue;
        if (debug >= 4) 
        {
            slog("debug: Running timeout %p (code=%p repeat=%d)",
                    to, to->code, to->repeat);
        }
        to->code(to);
        if (to->repeat) 
            continue;
        dlinkDelete(&to->node, &timeout_list);
        free(to);
    }
    if (debug >= 2)
        slog("debug: Finished timeout list");
}

/* Add a timeout to the list to be triggered in `delay' seconds.  If
 * `repeat' is nonzero, do not delete the timeout after it is triggered.
 * This must maintain the property that timeouts added from within a
 * timeout routine do not get checked during that run of the timeout list.
 */

struct Timeout *
add_timeout(int delay, void (*code)(struct Timeout *), int repeat)
{
    struct Timeout *t = smalloc(sizeof(struct Timeout));
    t->settime = time(NULL);
    t->timeout = t->settime + delay;
    t->code = code;
    t->repeat = repeat;
    dlinkAdd(t, &t->node, &timeout_list);
    return t;
}

/* Remove a timeout from the list (if it's there). */

void 
del_timeout(struct Timeout *t)
{
    dlinkFindDelete(&timeout_list, t);
    sfree(t);
}
